#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import copy
from typing import Any, List, Mapping, MutableMapping, Union

from airbyte_cdk.models import AirbyteStateBlob, AirbyteStateMessage, AirbyteStateType


class ConnectorStateManager:
    """
    ConnectorStateManager consolidates the various forms of a stream's incoming state message (STREAM / GLOBAL / LEGACY) under a common
    interface. It also provides methods to extract and update state
    """

    # In the immediate, we only persist legacy which will be used during abstract_source.read(). In the subsequent PRs we will
    # initialize the ConnectorStateManager according to the new per-stream interface received from the platform
    def __init__(self, state: Union[List[AirbyteStateMessage], MutableMapping[str, Any]] = None):
        if not state:
            self.legacy = {}
        elif self.is_migrated_legacy_state(state):
            # The legacy state format received from the platform is parsed and stored as a single AirbyteStateMessage when reading
            # the file. This is used for input backwards compatibility.
            self.legacy = state[0].data
        elif isinstance(state, MutableMapping):
            # In the event that legacy state comes in as its original JSON object format, no changes to the input need to be made
            self.legacy = state
        else:
            raise ValueError("Input state should come in the form of list of Airbyte state messages or a mapping of states")

    def get_stream_state(self, namespace: str, stream_name: str) -> AirbyteStateBlob:
        # todo implement in upcoming PRs
        pass

    def get_legacy_state(self) -> MutableMapping[str, Any]:
        """
        Returns a deep copy of the current legacy state dictionary made up of the state of all streams for a connector
        :return: A copy of the legacy state
        """
        return copy.deepcopy(self.legacy, {})

    def update_state_for_stream(self, namespace: str, stream_name: str, value: Mapping[str, Any]):
        # todo implement in upcoming PRs
        pass

    @staticmethod
    def is_migrated_legacy_state(state: List[AirbyteStateMessage]) -> bool:
        return (
            isinstance(state, List)
            and len(state) == 1
            and isinstance(state[0], AirbyteStateMessage)
            and state[0].type == AirbyteStateType.LEGACY
        )



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


import json
import logging
from abc import ABC, abstractmethod
from typing import Any, Generic, Iterable, List, Mapping, MutableMapping, TypeVar, Union

from airbyte_cdk.connector import BaseConnector, DefaultConnectorMixin, TConfig
from airbyte_cdk.models import AirbyteCatalog, AirbyteMessage, AirbyteStateMessage, AirbyteStateType, ConfiguredAirbyteCatalog

TState = TypeVar("TState")
TCatalog = TypeVar("TCatalog")


class BaseSource(BaseConnector[TConfig], ABC, Generic[TConfig, TState, TCatalog]):
    @abstractmethod
    def read_state(self, state_path: str) -> TState:
        ...

    @abstractmethod
    def read_catalog(self, catalog_path: str) -> TCatalog:
        ...

    @abstractmethod
    def read(self, logger: logging.Logger, config: TConfig, catalog: TCatalog, state: TState = None) -> Iterable[AirbyteMessage]:
        """
        Returns a generator of the AirbyteMessages generated by reading the source with the given configuration, catalog, and state.
        """

    @abstractmethod
    def discover(self, logger: logging.Logger, config: TConfig) -> AirbyteCatalog:
        """
        Returns an AirbyteCatalog representing the available streams and fields in this integration. For example, given valid credentials to a
        Postgres database, returns an Airbyte catalog where each postgres table is a stream, and each table column is a field.
        """


class Source(
    DefaultConnectorMixin,
    BaseSource[Mapping[str, Any], Union[List[AirbyteStateMessage], MutableMapping[str, Any]], ConfiguredAirbyteCatalog],
    ABC,
):
    # can be overridden to change an input state
    def read_state(self, state_path: str) -> List[AirbyteStateMessage]:
        """
        Retrieves the input state of a sync by reading from the specified JSON file. Incoming state can be deserialized into either
        a JSON object for legacy state input or as a list of AirbyteStateMessages for the per-stream state format. Regardless of the
        incoming input type, it will always be transformed and output as a list of AirbyteStateMessage(s).
        :param state_path: The filepath to where the stream states are located
        :return: The complete stream state based on the connector's previous sync
        """
        if state_path:
            state_obj = json.loads(open(state_path, "r").read())
            if not state_obj:
                return []
            is_per_stream_state = isinstance(state_obj, List)
            if is_per_stream_state:
                parsed_state_messages = []
                for state in state_obj:
                    parsed_message = AirbyteStateMessage.parse_obj(state)
                    if not parsed_message.stream and not parsed_message.data and not parsed_message.global_:
                        raise ValueError("AirbyteStateMessage should contain either a stream, global, or state field")
                    parsed_state_messages.append(parsed_message)
                return parsed_state_messages
            else:
                # Existing connectors that override read() might not be able to interpret the new state format. We temporarily
                # send state in the old format for these connectors, but once all have been upgraded, this block can be removed
                # vars(self.__class__) checks if the current class directly overrides the read() function
                if "read" in vars(self.__class__):
                    return state_obj
                return [AirbyteStateMessage(type=AirbyteStateType.LEGACY, data=state_obj)]
        return []

    # can be overridden to change an input catalog
    def read_catalog(self, catalog_path: str) -> ConfiguredAirbyteCatalog:
        return ConfiguredAirbyteCatalog.parse_obj(self.read_config(catalog_path))



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from contextlib import nullcontext as does_not_raise

import pytest
from airbyte_cdk.models import AirbyteStateMessage, AirbyteStateType
from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager


@pytest.mark.parametrize(
    "input_state, expected_legacy_state, expected_error",
    [
        pytest.param(
            [AirbyteStateMessage(type=AirbyteStateType.LEGACY, data={"actresses": {"id": "seehorn_rhea"}})],
            {"actresses": {"id": "seehorn_rhea"}},
            does_not_raise(),
            id="test_legacy_input_state",
        ),
        pytest.param(
            {
                "actors": {"created_at": "1962-10-22"},
                "actresses": {"id": "seehorn_rhea"},
            },
            {"actors": {"created_at": "1962-10-22"}, "actresses": {"id": "seehorn_rhea"}},
            does_not_raise(),
            id="test_supports_legacy_json_blob",
        ),
        pytest.param({}, {}, does_not_raise(), id="test_initialize_empty_mapping_by_default"),
        pytest.param([], {}, does_not_raise(), id="test_initialize_empty_state"),
        pytest.param("strings_are_not_allowed", None, pytest.raises(ValueError), id="test_value_error_is_raised_on_invalid_state_input"),
    ],
)
def test_get_legacy_state(input_state, expected_legacy_state, expected_error):
    with expected_error:
        state_manager = ConnectorStateManager(input_state)
        actual_legacy_state = state_manager.get_legacy_state()
        assert actual_legacy_state == expected_legacy_state



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, List, Mapping, Optional, Union

import requests
from airbyte_cdk.sources.declarative.decoders.decoder import Decoder
from airbyte_cdk.sources.declarative.decoders.json_decoder import JsonDecoder
from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.requesters.paginators.strategies.pagination_strategy import PaginationStrategy
from airbyte_cdk.sources.declarative.types import Config
from dataclasses_jsonschema import JsonSchemaMixin


@dataclass
class CursorPaginationStrategy(PaginationStrategy, JsonSchemaMixin):
    """
    Pagination strategy that evaluates an interpolated string to define the next page token

    Attributes:
        cursor_value (Union[InterpolatedString, str]): template string evaluating to the cursor value
        config (Config): connection config
        stop_condition (Optional[InterpolatedBoolean]): template string evaluating when to stop paginating
        decoder (Decoder): decoder to decode the response
    """

    cursor_value: Union[InterpolatedString, str]
    config: Config
    options: InitVar[Mapping[str, Any]]
    stop_condition: Optional[Union[InterpolatedBoolean, str]] = None
    decoder: Decoder = JsonDecoder(options={})

    def __post_init__(self, options: Mapping[str, Any]):
        if isinstance(self.cursor_value, str):
            self.cursor_value = InterpolatedString.create(self.cursor_value, options=options)
        if isinstance(self.stop_condition, str):
            self.stop_condition = InterpolatedBoolean(condition=self.stop_condition, options=options)

    def next_page_token(self, response: requests.Response, last_records: List[Mapping[str, Any]]) -> Optional[Any]:
        decoded_response = self.decoder.decode(response)

        # The default way that link is presented in requests.Response is a string of various links (last, next, etc). This
        # is not indexable or useful for parsing the cursor, so we replace it with the link dictionary from response.links
        headers = response.headers
        headers["link"] = response.links

        if self.stop_condition:
            should_stop = self.stop_condition.eval(self.config, response=decoded_response, headers=headers, last_records=last_records)
            if should_stop:
                return None
        token = self.cursor_value.eval(config=self.config, last_records=last_records, response=decoded_response, headers=headers)
        return token if token else None

    def reset(self):
        # No state to reset
        pass
